Documentation
¶
Overview ¶
Package errors provides simple error handling primitives including stack, annotations, and data.
Original package created by Dave Cheney Copyright (c) 2015, Dave Cheney <dave@cheney.net>
Modifications of the original package by Friends of Go Copyright (c) 2019, Friends of Go <contact@friendsofgo.tech>
Further modifications of the Friends of Go version by Adam Manwaring Copyright (c) 2019, Adam Manwaring <pantsmann@byu.net>
The traditional error handling idiom in Go is roughly akin to
if err != nil {
return err
}
which when applied recursively up the call stack results in error reports without context or debugging information. The errors package allows programmers to add context to the failure path in their code in a way that does not destroy the original value of the error.
Adding context to an error ¶
The errors.Wrap function returns a new error that adds context to the original error by recording a stack trace at the point Wrap is called, together with the supplied message. For example
_, err := ioutil.ReadAll(r)
if err != nil {
return errors.Wrap(err, "read failed")
}
If additional control is required, the errors.WithStack and errors.WithMessage functions destructure errors.Wrap into its component operations: annotating an error with a stack trace and with a message, respectively.
Retrieving the cause of an error ¶
Using errors.Wrap constructs a stack of errors, adding context to the preceding error. Depending on the nature of the error it may be necessary to reverse the operation of errors.Wrap to retrieve the original error for inspection. Any error value which implements this interface
type causer interface {
Cause() error
}
can be inspected by errors.Cause. errors.Cause will recursively retrieve the topmost error that does not implement causer, which is assumed to be the original cause. For example:
switch err := errors.Cause(err).(type) {
case *MyError:
// handle specifically
default:
// unknown error
}
Although the causer interface is not exported by this package, it is considered a part of its stable public interface.
With the new standard package error we have two new ways to figure what is the cause of our error:
var target *MyError
if errors.As(err, &target) {
// handle specifically
} else {
// unknown error
}
or even with sentinel errors:
var ErrMyError = errors.New("my sentinel error")
if errors.Is(err, ErrMyError) {
// handle specifically
} else {
// unknown error
}
Formatted printing of errors ¶
All error values returned from this package implement fmt.Formatter and can be formatted by the fmt package. The following verbs are supported:
%s print the error. If the error is a wrapper it will be
printed recursively.
%v see %s
%+v extended format. Each Frame of the error's StackTrace will
be printed in detail. Any key/value pairs recorded WithData()
will also be printed.
Retrieving the stack trace of an error or wrapper ¶
New, Errorf, Wrap, and Wrapf record a stack trace at the point they are invoked. This information can be retrieved with the following interface:
type stackTracer interface {
StackTrace() errors.StackTrace
}
The returned errors.StackTrace type is defined as
type StackTrace []Frame
The Frame type represents a call site in the stack trace. Frame supports the fmt.Formatter interface that can be used for printing information about the stack trace of this error. For example:
if err, ok := err.(stackTracer); ok {
for _, f := range err.StackTrace() {
fmt.Printf("%+s:%d\n", f, f)
}
}
Although the stackTracer interface is not exported by this package, it is considered a part of its stable public interface.
See the documentation for Frame.Format for more details.
Retrieving data from an error chain ¶
Key/value pairs recorded (using WithData or WrapWithData) in errors of the error chain can be retrieved with the following interface:
type dataCacher interface {
DataCache() map[string]interface{}
}
The returned map will contain all the key/value pairs from any dataCachers of equal or greater depth in the error chain. For example (assuming the dataError definition above):
var d dataError
if errors.As(err, &d) {
kv := d.DataCache()
}
Note that if more than one error in the chain contains the same key, DataCache() returns only the shallowest value for that key. That is:
err = errors.WithData(err, "deep", 1, "dup", 1)
kv := errors.WithData(err, "dup" 2, "shallow", 2).DataCache()
fmt.Printf("key/value pairs: %v", kv)
// Example Output:
// key/value pairs: map[deep:1 dup:2 shallow:2]
The when using %+v to format the error all keys and values are output at the level they were set.
Example (StackTrace) ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func fn() error {
e1 := errors.New("error")
e2 := errors.Wrap(e1, "inner")
e3 := errors.Wrap(e2, "middle")
return errors.Wrap(e3, "outer")
}
func main() {
type stackTracer interface {
StackTrace() errors.StackTrace
}
err, ok := errors.Cause(fn()).(stackTracer)
if !ok {
panic("oops, err does not implement stackTracer")
}
st := err.StackTrace()
fmt.Printf("%+v", st[0:2]) // top two frames
// Example output:
// github.com/pantsmann/errors_test.fn
// /home/dfc/src/github.com/pantsmann/errors/example_test.go:47
// github.com/pantsmann/errors_test.Example_stackTrace
// /home/dfc/src/github.com/pantsmann/errors/example_test.go:127
}
Output:
Index ¶
- func As(err error, target interface{}) bool
- func Cause(err error) error
- func Errorf(format string, args ...interface{}) error
- func Is(err error, target error) bool
- func New(message string) error
- func Unwrap(err error) error
- func WithData(err error, keyVals ...interface{}) error
- func WithMessage(err error, message string) error
- func WithMessagef(err error, format string, args ...interface{}) error
- func WithStack(err error) error
- func Wrap(err error, message string) error
- func WrapWithData(err error, message string, keyVals ...interface{}) error
- func Wrapf(err error, format string, args ...interface{}) error
- type Annotatable
- type Base
- type Frame
- type StackTrace
- type Wrapper
Examples ¶
Constants ¶
This section is empty.
Variables ¶
This section is empty.
Functions ¶
func As ¶ added in v0.9.3
As finds the first error in err's chain that matches target, and if so, sets target to that error value and returns true.
The chain consists of err itself followed by the sequence of errors obtained by repeatedly calling Unwrap.
An error matches target if the error's concrete value is assignable to the value pointed to by target, or if the error has a method As(interface{}) bool such that As(target) returns true. In the latter case, the As method is responsible for setting target.
As will panic if target is not a non-nil pointer to either a type that implements error, or to any interface type. As returns false if err is nil.
func Cause ¶
Cause returns the underlying cause of the error, if possible. An error value has a cause if it implements the standard errors.Wrapper interface:
type Wrapper interface {
Unwrap() error
}
If the error does not implement Wrapper, the original error will be returned. If the error is nil, nil will be returned without further investigation.
Example ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func fn() error {
e1 := errors.New("error")
e2 := errors.Wrap(e1, "inner")
e3 := errors.Wrap(e2, "middle")
return errors.Wrap(e3, "outer")
}
func main() {
err := fn()
fmt.Println(err)
fmt.Println(errors.Cause(err))
}
Output: outer: middle: inner: error error
Example (Printf) ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func main() {
err := errors.Wrap(func() error {
return func() error {
return errors.Errorf("hello %s", fmt.Sprintf("world"))
}()
}(), "failed")
fmt.Printf("%v", err)
}
Output: failed: hello world
func Errorf ¶ added in v0.3.0
Errorf formats according to a format specifier and returns the string as a value that satisfies error. Errorf also records the stack trace at the point it was called.
Example (Extended) ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func main() {
err := errors.Errorf("whoops: %s", "foo")
fmt.Printf("%+v", err)
// Example output:
// whoops: foo
// github.com/pantsmann/errors_test.ExampleErrorf
// /home/dfc/src/github.com/pantsmann/errors/example_test.go:101
// testing.runExample
// /home/dfc/go/src/testing/example.go:114
// testing.RunExamples
// /home/dfc/go/src/testing/example.go:38
// testing.(*M).Run
// /home/dfc/go/src/testing/testing.go:744
// main.main
// /github.com/pantsmann/errors/_test/_testmain.go:102
// runtime.main
// /home/dfc/go/src/runtime/proc.go:183
// runtime.goexit
// /home/dfc/go/src/runtime/asm_amd64.s:2059
}
Output:
func Is ¶ added in v0.9.3
Is reports whether any error in err's chain matches target.
The chain consists of err itself followed by the sequence of errors obtained by repeatedly calling Unwrap.
An error is considered to match a target if it is equal to that target or if it implements a method Is(error) bool such that Is(target) returns true.
func New ¶
New returns an error with the supplied message. New also records the stack trace at the point it was called.
Example ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func main() {
err := errors.New("whoops")
fmt.Println(err)
}
Output: whoops
Example (Printf) ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func main() {
err := errors.New("whoops")
fmt.Printf("%+v", err)
// Example output:
// whoops
// github.com/pantsmann/errors_test.ExampleNew_printf
// /home/dfc/src/github.com/pantsmann/errors/example_test.go:17
// testing.runExample
// /home/dfc/go/src/testing/example.go:114
// testing.RunExamples
// /home/dfc/go/src/testing/example.go:38
// testing.(*M).Run
// /home/dfc/go/src/testing/testing.go:744
// main.main
// /github.com/pantsmann/errors/_test/_testmain.go:106
// runtime.main
// /home/dfc/go/src/runtime/proc.go:183
// runtime.goexit
// /home/dfc/go/src/runtime/asm_amd64.s:2059
}
Output:
func Unwrap ¶ added in v0.9.3
Unwrap returns the result of calling the Unwrap method on err, if err implements Unwrap. Otherwise, Unwrap returns nil.
func WithData ¶ added in v0.9.3
WithData annotates err with a map of key/value pairs. keyVals should be passed in as pairs; the first of each pair being a string (the key). If an odd number of keyVals are passed in, the last one is ignored. If a key value is not a string, it and the next keyVal element (the value) are skipped. If err is nil, WithData returns nil.
func WithMessage ¶ added in v0.8.0
WithMessage annotates err with a new message. If err is nil, WithMessage returns nil.
Example ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func main() {
cause := errors.New("whoops")
err := errors.WithMessage(cause, "oh noes")
fmt.Println(err)
}
Output: oh noes: whoops
func WithMessagef ¶ added in v0.8.1
WithMessagef annotates err with the format specifier. If err is nil, WithMessagef returns nil.
func WithStack ¶ added in v0.8.0
WithStack annotates err with a stack trace at the point WithStack was called. If err is nil, WithStack returns nil.
Example ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func main() {
cause := errors.New("whoops")
err := errors.WithStack(cause)
fmt.Println(err)
}
Output: whoops
Example (Printf) ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func main() {
cause := errors.New("whoops")
err := errors.WithStack(cause)
fmt.Printf("%+v", err)
// Example Output:
// whoops
// github.com/pantsmann/errors_test.ExampleWithStack_printf
// /home/fabstu/go/src/github.com/pantsmann/errors/example_test.go:55
// testing.runExample
// /usr/lib/go/src/testing/example.go:114
// testing.RunExamples
// /usr/lib/go/src/testing/example.go:38
// testing.(*M).Run
// /usr/lib/go/src/testing/testing.go:744
// main.main
// github.com/pantsmann/errors/_test/_testmain.go:106
// runtime.main
// /usr/lib/go/src/runtime/proc.go:183
// runtime.goexit
// /usr/lib/go/src/runtime/asm_amd64.s:2086
// github.com/pantsmann/errors_test.ExampleWithStack_printf
// /home/fabstu/go/src/github.com/pantsmann/errors/example_test.go:56
// testing.runExample
// /usr/lib/go/src/testing/example.go:114
// testing.RunExamples
// /usr/lib/go/src/testing/example.go:38
// testing.(*M).Run
// /usr/lib/go/src/testing/testing.go:744
// main.main
// github.com/pantsmann/errors/_test/_testmain.go:106
// runtime.main
// /usr/lib/go/src/runtime/proc.go:183
// runtime.goexit
// /usr/lib/go/src/runtime/asm_amd64.s:2086
}
Output:
func Wrap ¶
Wrap returns an error annotating err with a stack trace at the point Wrap is called, and the supplied message. If err is nil, Wrap returns nil.
Example ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func main() {
cause := errors.New("whoops")
err := errors.Wrap(cause, "oh noes")
fmt.Println(err)
}
Output: oh noes: whoops
Example (Extended) ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func fn() error {
e1 := errors.New("error")
e2 := errors.Wrap(e1, "inner")
e3 := errors.Wrap(e2, "middle")
return errors.Wrap(e3, "outer")
}
func main() {
err := fn()
fmt.Printf("%+v\n", err)
// Example output:
// error
// github.com/pantsmann/errors_test.fn
// /home/dfc/src/github.com/pantsmann/errors/example_test.go:47
// github.com/pantsmann/errors_test.ExampleCause_printf
// /home/dfc/src/github.com/pantsmann/errors/example_test.go:63
// testing.runExample
// /home/dfc/go/src/testing/example.go:114
// testing.RunExamples
// /home/dfc/go/src/testing/example.go:38
// testing.(*M).Run
// /home/dfc/go/src/testing/testing.go:744
// main.main
// /github.com/pantsmann/errors/_test/_testmain.go:104
// runtime.main
// /home/dfc/go/src/runtime/proc.go:183
// runtime.goexit
// /home/dfc/go/src/runtime/asm_amd64.s:2059
// github.com/pantsmann/errors_test.fn
// /home/dfc/src/github.com/pantsmann/errors/example_test.go:48: inner
// github.com/pantsmann/errors_test.fn
// /home/dfc/src/github.com/pantsmann/errors/example_test.go:49: middle
// github.com/pantsmann/errors_test.fn
// /home/dfc/src/github.com/pantsmann/errors/example_test.go:50: outer
}
Output:
func WrapWithData ¶ added in v0.9.3
WrapWithData returns an error annotating err with a stack trace at the point WrapWithData is called, the supplied message, and any supplied key/value pairs. If err is nil, WrapWithData returns nil.
func Wrapf ¶ added in v0.2.0
Wrapf returns an error annotating err with a stack trace at the point Wrapf is called, and the format specifier. If err is nil, Wrapf returns nil.
Example ¶
package main
import (
"fmt"
"github.com/pantsmann/errors"
)
func main() {
cause := errors.New("whoops")
err := errors.Wrapf(cause, "oh noes #%d", 2)
fmt.Println(err)
}
Output: oh noes #2: whoops
Types ¶
type Annotatable ¶ added in v0.9.3
Annotatable represents errors that are unwrappable and formatable. Errors that implement special behaviors should implement this interface either directly or through embedding in order to avoid breaking Wrap() semantics. As a simple example, given the following definitions:
type mySpecial struct { errors.Annotatable }
func (m mySpecial) Special() bool { return true }
the following code would add the Special() behavior to a wrapped error
specialErr := mySpecial{errors.Wrap(someGenericError, "You are now a special error").(errors.Annotatable)}
type Base ¶ added in v0.9.3
type Base struct{ Err error }
Base is an error with basic annotatable functionality. It can be used to simply add Annotatable to special error behaviors by extending an embedded error to fulfill Annotatable. For example given the following definitions:
type mySpecial struct { errors.Annotatable }
func (m mySpecial) Special() bool { return true }
the following would add Special() as a behavior to someGenericError and this would be findable using errors.As and one would still be able to output stack traces inside of someGenericError (if they exist).
specialErr := mySpecial{errors.Base{someGenericError}}
type Frame ¶ added in v0.6.0
type Frame uintptr
Frame represents a program counter inside a stack frame. For historical reasons if Frame is interpreted as a uintptr its value represents the program counter + 1.
func (Frame) Format ¶ added in v0.6.0
Format formats the frame according to the fmt.Formatter interface.
%s source file %d source line %n function name %v equivalent to %s:%d
Format accepts flags that alter the printing of some verbs, as follows:
%+s function name and path of source file relative to the compile time
GOPATH separated by \n\t (<funcname>\n\t<path>)
%+v equivalent to %+s:%d
func (Frame) MarshalText ¶ added in v0.9.3
MarshalText formats a stacktrace Frame as a text string. The output is the same as that of fmt.Sprintf("%+v", f), but without newlines or tabs.
type StackTrace ¶ added in v0.7.0
type StackTrace []Frame
StackTrace is stack of Frames from innermost (newest) to outermost (oldest).
func (StackTrace) Format ¶ added in v0.7.0
func (st StackTrace) Format(s fmt.State, verb rune)
Format formats the stack of Frames according to the fmt.Formatter interface.
%s lists source files for each Frame in the stack %v lists the source file and line number for each Frame in the stack
Format accepts flags that alter the printing of some verbs, as follows:
%+v Prints filename, function, and line number for each Frame in the stack.